import * as vscode from 'vscode';
import axios from 'axios';

let config: {
    apiEndpoint: string;
    apiAuthentication: string;
    apiModel: string;
    apiTemperature: number;
    numPredict: number;
    promptWindowSize: number;
    completionKeys: string[];
    responsePreview: boolean;
    responsePreviewMaxTokens: number;
    responsePreviewDelay: number;
    continueInline: boolean;
    keepAlive: number;
    topP: number;
};

let previewDecorationType: vscode.TextEditorDecorationType;
let activeCompletionManager: CompletionManager | null = null;

function updateConfig() {
    const vsConfig = vscode.workspace.getConfiguration('fabelous-autocoder');
    config = {
        apiEndpoint: vsConfig.get('endpoint') || 'http://localhost:11434/api/generate',
        apiAuthentication: vsConfig.get('authentication') || '',
        apiModel: vsConfig.get('model') || 'fabelous-coder:latest',
        apiTemperature: vsConfig.get('temperature') || 0.7,
        numPredict: vsConfig.get('max tokens predicted') || 1000,
        promptWindowSize: vsConfig.get('prompt window size') || 2000,
        completionKeys: (vsConfig.get('completion keys') as string || ' ').split(''),
        responsePreview: vsConfig.get('response preview') || false,
        responsePreviewMaxTokens: vsConfig.get('preview max tokens') || 50,
        responsePreviewDelay: vsConfig.get('preview delay') || 0,
        continueInline: vsConfig.get('continue inline') || false,
        keepAlive: vsConfig.get('keep alive') || 30,
        topP: vsConfig.get('top p') || 1,
    };
}

function createPreviewDecorationType() {
    previewDecorationType = vscode.window.createTextEditorDecorationType({
        after: {
            color: '#888888',
            fontStyle: 'italic',
        },
        textDecoration: 'none; display: none;',
    });
}

function getContextLines(document: vscode.TextDocument, position: vscode.Position): string {
    const startLine = Math.max(0, position.line - 1);
    const endLine = position.line;
    return document.getText(new vscode.Range(startLine, 0, endLine, position.character));
}

function createFIMPrompt(prefix: string, language: string): string {
    return `<fim_prefix>${prefix}<fim_middle><fim_suffix>${language}\n`;
}

async function generateCompletion(prompt: string, cancellationToken: vscode.CancellationToken): Promise<string> {
    const axiosCancelToken = new axios.CancelToken((c) => {
        cancellationToken.onCancellationRequested(() => c('Request cancelled'));
    });

    const response = await axios.post(config.apiEndpoint, {
        model: config.apiModel,
        prompt: prompt,
        stream: false,
        raw: true,
        options: {
            num_predict: config.numPredict,
            temperature: config.apiTemperature,
            stop: ['<fim_suffix>', '```'],
            keep_alive: config.keepAlive,
            top_p: config.topP,
        }
    }, {
        cancelToken: axiosCancelToken,
        headers: {
            'Authorization': config.apiAuthentication
        }
    });

    return response.data.response.replace(/<fim_middle>|<fim_suffix>|<fim_prefix>/g, '').trim();
}
class CompletionManager {
    private textEditor: vscode.TextEditor;
    private document: vscode.TextDocument;
    private startPosition: vscode.Position;
    private completionText: string;

    constructor(textEditor: vscode.TextEditor, startPosition: vscode.Position, completionText: string) {
        this.textEditor = textEditor;
        this.document = textEditor.document;
        this.startPosition = startPosition;
        this.completionText = completionText;
    }

    public async showPreview() {
        // Calculate lines in the completionText
        const completionLines = this.completionText.split('\n').length;

        // Insert empty lines to make space for completion preview
        const edit = new vscode.WorkspaceEdit();
        const linePadding = '\n'.repeat(completionLines); // Adjust padding based on lines in completion
        const range = new vscode.Range(this.startPosition, this.startPosition);
        
        edit.insert(this.document.uri, this.startPosition, linePadding);
        await vscode.workspace.applyEdit(edit);

        const previewRanges: vscode.DecorationOptions[] = this.completionText.split('\n').map((line, index) => {
            const actualLineNumber = this.startPosition.line + index;
            return {
                range: new vscode.Range(
                    new vscode.Position(actualLineNumber, 0),
                    new vscode.Position(actualLineNumber, 0) // Positions should be zero-length for insertions
                ),
                renderOptions: {
                    after: {
                        contentText: ` ${line}`.trim(),
                    },
                },
            };
        });

        this.textEditor.setDecorations(previewDecorationType, previewRanges);
    }

    public async acceptCompletion() {
        const edit = new vscode.WorkspaceEdit();

        // Prepare to insert completion text
        const endLine = this.startPosition.line + this.completionText.split('\n').length;

        // Replace the preview lines with actual completion text
        const rangeToReplace = new vscode.Range(
            this.startPosition,
            new vscode.Position(endLine, 0)  // Position for line end should just reach intended insert
        );

        edit.replace(this.document.uri, rangeToReplace, this.completionText);

        await vscode.workspace.applyEdit(edit);

        this.clearPreview();
    }

    public clearPreview() {
        this.textEditor.setDecorations(previewDecorationType, []); // Clear only the decorations
    }

    public declineCompletion() {
        this.clearPreview();
    }
}


async function autocompleteCommand(textEditor: vscode.TextEditor, edit: vscode.TextEditorEdit, ...args: any[]) {
    const cancellationTokenSource = new vscode.CancellationTokenSource();
    const cancellationToken = cancellationTokenSource.token;

    try {
        const document = textEditor.document;
        const position = textEditor.selection.active;
        const context = getContextLines(document, position);
        const fimPrompt = createFIMPrompt(context, document.languageId);

        const completionText = await vscode.window.withProgress({
            location: vscode.ProgressLocation.Notification,
            title: 'Fabelous Autocoder',
            cancellable: true,
        }, async (progress, progressCancellationToken) => {
            progress.report({ message: 'Generating...' });
            return await generateCompletion(fimPrompt, progressCancellationToken);
        });

        console.log('Completion generated:', completionText);


        const completionManager = new CompletionManager(textEditor, position, completionText);
        await completionManager.showPreview();
        activeCompletionManager = completionManager;

        let isDisposed = false;

        const dispose = () => {
            if (!isDisposed) {
                console.log('Disposing listeners');
                disposable.dispose();
                typeDisposable.dispose();
                activeCompletionManager = null;
                isDisposed = true;
            }
        };

        const disposable = vscode.Disposable.from(
            vscode.window.onDidChangeTextEditorSelection(async (event) => {
                if (event.textEditor !== textEditor) return;

                if (event.kind === vscode.TextEditorSelectionChangeKind.Keyboard) {
                    console.log('Accepting completion');
                    await completionManager.acceptCompletion();
                    dispose();
                }
            }),
            vscode.window.onDidChangeActiveTextEditor(() => {
                console.log('Active editor changed, clearing preview');
                completionManager.clearPreview();
                dispose();
            }),
            vscode.workspace.onDidChangeTextDocument((event) => {
                if (event.document === document) {
                    console.log('Document changed, clearing preview');
                    completionManager.clearPreview();
                    dispose();
                }
            })
        );

        const typeDisposable = vscode.commands.registerCommand('type', async (args) => {
            if (args.text === '\b') { // Backspace key
                console.log('Declining completion');
                completionManager.declineCompletion();
                dispose();
            }
        });

    } catch (err: any) {
        console.error('Error in autocompleteCommand:', err);
        vscode.window.showErrorMessage(`Fabelous Autocoder encountered an error: ${err.message}`);
    } finally {
        cancellationTokenSource.dispose();
    }
}

async function acceptCompletion() {
    if (activeCompletionManager) {
        await activeCompletionManager.acceptCompletion();
        const editor = vscode.window.activeTextEditor;
        if (editor) {
            const lastLine = editor.document.lineAt(editor.document.lineCount - 1);
            const newPosition = new vscode.Position(lastLine.lineNumber, lastLine.text.length);
            editor.selection = new vscode.Selection(newPosition, newPosition);
        }
        activeCompletionManager = null;
    }
}

async function handleTab() {
    if (activeCompletionManager) {
        await acceptCompletion();
    } else {
        await vscode.commands.executeCommand('tab');
    }
}

async function provideCompletionItems(document: vscode.TextDocument, position: vscode.Position, cancellationToken: vscode.CancellationToken) {
    const item = new vscode.CompletionItem('Fabelous autocompletion');
    item.insertText = new vscode.SnippetString('${1:}');
    item.documentation = new vscode.MarkdownString('Press `Enter` to get an autocompletion from Fabelous Autocoder');

    if (config.responsePreview) {
        await new Promise(resolve => setTimeout(resolve, config.responsePreviewDelay * 1000));
        if (cancellationToken.isCancellationRequested) {
            return [item];
        }

        const context = getContextLines(document, position);
        const fimPrompt = createFIMPrompt(context, document.languageId);

        try {
            const result = await generateCompletion(fimPrompt, cancellationToken);
            const preview = (result as any).preview;
            if (preview) {
                item.detail = preview.split('\n')[0];
            }
        } catch (error) {
            console.error('Error fetching preview:', error);
        }
    }

    if (config.continueInline || !config.responsePreview) {
        item.command = {
            command: 'fabelous-autocoder.autocomplete',
            title: 'Fabelous Autocomplete',
            arguments: []
        };
    }

    return [item];
}
export function activate(context: vscode.ExtensionContext) {
    updateConfig();
    createPreviewDecorationType();

    context.subscriptions.push(
        vscode.workspace.onDidChangeConfiguration(updateConfig),
        vscode.languages.registerCompletionItemProvider('*', { provideCompletionItems }, ...config.completionKeys),
        vscode.commands.registerTextEditorCommand('fabelous-autocoder.autocomplete', autocompleteCommand),
        vscode.commands.registerCommand('fabelous-autocoder.acceptCompletion', acceptCompletion),
        vscode.commands.registerCommand('fabelous-autocoder.handleTab', handleTab)
    );
}


export function deactivate() {}
